(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('astring'), require('periscopic'), require('acorn'), require('estree-walker')) :
	typeof define === 'function' && define.amd ? define(['exports', 'astring', 'periscopic', 'acorn', 'estree-walker'], factory) :
	(factory((global.CodeRed = {}),global.astring,global.perisopic,global.acorn,global.estreeWalker));
}(this, (function (exports,astring,perisopic,acorn,estreeWalker) { 'use strict';

	function deconflict(name, names) {
		const original = name;
		let i = 1;

		while (names.has(name)) {
			name = `${original}$${i++}`;
		}

		return name;
	}

	function print(node, opts = {}) {
		const {
			getName = (x) => x
		} = opts;

		const { map: scope_map } = perisopic.analyze(node);
		const deconflicted = new WeakMap();

		const generator = Object.assign({}, astring.baseGenerator, {
			handle( node, state) {
				if (Array.isArray(node)) {
					console.log(node);
					throw new Error('hmm');
					for (let i = 0; i < node.length; i += 1) {
						this.handle(node[i], state);
						if (i < node.length - 1) {
							state.write(state.lineEnd);
							state.write(state.indent);
						}
					}

					return;
				}

				if (!node.type) {
					console.log(`missing type: `, node);
				}

				if (!this[node.type]) {
					throw new Error(`Not implemented: ${node.type}`);
				}

				try {
					this[node.type](node, state);
				} catch (err) {
					if (!err.depth) {
						console.log(`${err.message} while handling`, JSON.stringify(node, null, '  '));
						err.depth = 1;
					} else if (err.depth <= 3) {
						console.log(`${err.depth}:`, JSON.stringify(node, null, '  '));
						err.depth += 1;
					}

					throw err;
				}
			},

			AwaitExpression( node, state) {
				state.write('await ');
				const { argument } = node;
				this[argument.type](argument, state);
			},

			Identifier( node, state) {
				if (!node.name) {
					console.log(node);
				}

				if (node.name[0] === '@') {
					node = { ...node, name: getName(node.name.slice(1)) };
				}

				if (node.name[0] === '#') {
					const scope = scope_map.get(node);
					if (!scope) {
						throw new Error(`Could not find scope for node`);
					}
					const owner = scope.find_owner(node.name);

					if (!deconflicted.has(owner)) {
						deconflicted.set(owner, new Map());
					}

					const deconflict_map = deconflicted.get(owner);

					if (!deconflict_map.has(node.name)) {
						deconflict_map.set(node.name, deconflict(node.name.slice(1), owner.references));
					}

					const name = deconflict_map.get(node.name);
					node = { ...node, name };
				}

				return astring.baseGenerator.Identifier.call(this, node, state);
			},

			Literal( node, state) {
				if (typeof node.value === 'string') {
					state.write(JSON.stringify(node.value));
					return;
				}

				astring.baseGenerator.Literal.call(this, node, state);
			}
		});

		// const map = new SourceMap.SourceMapGenerator({
		// 	file: opts.file
		// });

		const code = astring.generate(node , {
			indent: '\t',
			generator
		});

		return {
			code,
			map: null 
		};
	}

	// generate an ID that is, to all intents and purposes, unique
	const id = (Math.round(Math.random() * 1e20)).toString(36);
	const re = new RegExp(`_${id}_(?:(\\d+)|(AT)|(HASH))_(\\w+)?`, 'g');

	const sigils = {
		'@': 'AT',
		'#': 'HASH'
	};

	const join = (strings) => {
		let str = strings[0];
		for (let i = 1; i < strings.length; i += 1) {
			str += `_${id}_${i - 1}_${strings[i]}`;
		}
		return str.replace(/([@#])(\w+)/g, (_m, sigil, name) => `_${id}_${sigils[sigil]}_${name}`);
	};

	const flatten_body = (array, target) => {
		for (let i = 0; i < array.length; i += 1) {
			const statement = array[i];
			if (Array.isArray(statement)) {
				flatten_body(statement, target);
				continue;
			}

			if (statement.type === 'ExpressionStatement') {
				if (!statement.expression) continue;

				if (Array.isArray(statement.expression)) {
					flatten_body(statement.expression, target);
					continue;
				}
			}

			target.push(statement);
		}

		return target;
	};

	const flatten = (nodes, target) => {
		for (let i = 0; i < nodes.length; i += 1) {
			const node = nodes[i];

			if (!node) continue;

			if (Array.isArray(node)) {
				flatten(node, target);
				continue;
			}

			target.push(node);
		}

		return target;
	};

	const inject = (node, values) => {
		estreeWalker.walk(node, {
			leave(node, parent, key, index) {
				delete node.start;
				delete node.end;

				if (node.type === 'Identifier') {
					re.lastIndex = 0;
					const match = re.exec(node.name);

					if (match) {
						if (match[1]) {
							if (+match[1] in values) {
								const value = values[+match[1]];

								if (index === null) {
									parent[key] = value;
								} else {
									parent[key][index] = value;
								}
							}
						} else {
							node.name = `${match[2] ? `@` : `#`}${match[4]}`;
						}
					}
				}

				if (node.type === 'Literal') {
					if (typeof node.value === 'string') {
						re.lastIndex = 0;
						node.value = node.value.replace(re, (m, i) => +i in values ? values[+i] : m);
					}
				}

				if (node.type === 'Program' || node.type === 'BlockStatement') {
					node.body = flatten_body(node.body, []);
				}

				if (node.params) node.params = flatten(node.params, []);
				if (node.arguments) node.arguments = flatten(node.arguments, []);
				if (node.specifiers) node.specifiers = flatten(node.specifiers, []);
			}
		});
	};

	function b(strings, ...values) {
		const str = join(strings);
		try {
			const ast = acorn.parse(str, {
				ecmaVersion: 2019,
				sourceType: 'module',
				allowAwaitOutsideFunction: true,
				allowReturnOutsideFunction: true
			});

			inject(ast, values);

			return ast.body;
		} catch (err) {
			console.log(`failed to parse:\n${str}`); // TODO proper error reporting
			throw err;
		}
	}

	function x(strings, ...values) {
		const str = join(strings);

		try {
			const expression = acorn.parseExpressionAt(str, 0, {
				allowAwaitOutsideFunction: true,
				allowImportExportEverywhere: true,
				allowReturnOutsideFunction: true
			});

			inject(expression, values);

			return expression;
		} catch (err) {
			console.log(`failed to parse:\n${str}`); // TODO proper error reporting
			throw err;
		}
	}

	exports.b = b;
	exports.x = x;
	exports.print = print;

	Object.defineProperty(exports, '__esModule', { value: true });

})));
